/* tcp-web-client.cc
// TE816 - Modelagem e Avaliação de Desempenho
// Pós Graduação em Engenharia Elétrica
// Universidade Federal do Paraná
// This souce was developed as basis for homework 
//
// -- 10/11/2018
// ==================================
// Prof. Carlos Marcelo Pedroso
// pedroso@eletrica.ufpr.br
// http://www.eletrica.ufpr.br/pedroso
// Universidade Federal do Paraná
// Departamento de Engenharia Elétrica
// Centro Politécnico
// CEP: 81531-990 - Curitiba - PR - Brasil
 */

#include "ns3/log.h"
#include "ns3/ipv4-address.h"
#include "ns3/ipv6-address.h"
#include "ns3/address-utils.h"
#include "ns3/nstime.h"
#include "ns3/inet-socket-address.h"
#include "ns3/inet6-socket-address.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/uinteger.h"
#include "ns3/trace-source-accessor.h"
#include "tcp-web-client.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE ("TcpWebClientApplication");
NS_OBJECT_ENSURE_REGISTERED (TcpWebClient);

TypeId
TcpWebClient::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::TcpWebClient")
    .SetParent<Application> ()
    .AddConstructor<TcpWebClient> ()
    .AddAttribute ("MaxRequests",
                   "The maximum number of packets the application will send",
                   UintegerValue (100),
                   MakeUintegerAccessor (&TcpWebClient::m_count),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("Interval",
                   "The time to wait between requests",
                   TimeValue (Seconds (1.0)),
                   MakeTimeAccessor (&TcpWebClient::m_interval),
                   MakeTimeChecker ())
    .AddAttribute ("RemoteAddress",
                   "The destination address of the outbound packets",
                   AddressValue (),
                   MakeAddressAccessor (&TcpWebClient::m_peerAddress),
                   MakeAddressChecker ())
    .AddAttribute ("RemotePort",
                   "The destination port of the outbound packets",
                   UintegerValue (0),
                   MakeUintegerAccessor (&TcpWebClient::m_peerPort),
                   MakeUintegerChecker<uint16_t> ())
    .AddAttribute ("RequestSize", "Size of request to be sent do server",
                   UintegerValue (100),
                   MakeUintegerAccessor (&TcpWebClient::SetDataSize,
                                         &TcpWebClient::GetDataSize),
                   MakeUintegerChecker<uint32_t> ())
    .AddTraceSource ("Tx", "A new packet is created and is sent",
                     MakeTraceSourceAccessor (&TcpWebClient::m_txTrace),
                     "ns3::Packet::TracedCallback")

  ;
  return tid;
}

TcpWebClient::TcpWebClient ()
{
  NS_LOG_FUNCTION_NOARGS ();
  m_sent = 0;
  m_bytesSent = 0;
  m_recvBack = 0;
  m_bytesRecvBack = 0;
  m_socket = 0;
  m_sendEvent = EventId ();
  m_data = 0;
  m_dataSize = 0;
}

TcpWebClient::~TcpWebClient()
{
  NS_LOG_FUNCTION_NOARGS ();
  m_socket = 0;

  delete [] m_data;
  m_data = 0;
  m_dataSize = 0;
}

void
TcpWebClient::SetRemote (Address ip, uint16_t port)
{
  m_peerAddress = ip;
  m_peerPort = port;
}

void
TcpWebClient::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  Application::DoDispose ();
}

void
TcpWebClient::StartApplication (void)
{
  NS_LOG_FUNCTION_NOARGS ();

  if (m_socket == 0)
    {
      TypeId tid = TypeId::LookupByName ("ns3::TcpSocketFactory");
      m_socket = Socket::CreateSocket (GetNode (), tid);
      if (Ipv4Address::IsMatchingType(m_peerAddress) == true)
        {
          m_socket->Bind();
          m_socket->Connect (InetSocketAddress (Ipv4Address::ConvertFrom(m_peerAddress), m_peerPort));
        }
      else if (Ipv6Address::IsMatchingType(m_peerAddress) == true)
        {
          m_socket->Bind6();
          m_socket->Connect (Inet6SocketAddress (Ipv6Address::ConvertFrom(m_peerAddress), m_peerPort));
        }
    }

  m_socket->SetRecvCallback (MakeCallback (&TcpWebClient::ReceivePacket, this));

  ScheduleTransmit (Seconds (0.));
}

void
TcpWebClient::StopApplication ()
{
  NS_LOG_FUNCTION_NOARGS ();

  if (m_socket != 0)
    {
      m_socket->Close ();
      m_socket->SetRecvCallback (MakeNullCallback<void, Ptr<Socket> > ());
      m_socket = 0;
    }

  Simulator::Cancel (m_sendEvent);
}

void
TcpWebClient::SetDataSize (uint32_t dataSize)
{
  NS_LOG_FUNCTION (dataSize);

  //
  // If the client is setting the request size this way, 
  // he doesn't care about the contents of the request at all, so
  // neither will we.
  //
  delete [] m_data;
  m_data = 0;
  m_dataSize = 0;
  m_size = dataSize;
}

uint32_t
TcpWebClient::GetDataSize (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_size;
}

void
TcpWebClient::SetFill (std::string fill)
{
  NS_LOG_FUNCTION (fill);

  uint32_t dataSize = fill.size () + 1;

  if (dataSize != m_dataSize)
    {
      delete [] m_data;
      m_data = new uint8_t [dataSize];
      m_dataSize = dataSize;
    }

  memcpy (m_data, fill.c_str (), dataSize);

  //
  // Overwrite packet size attribute.
  //
  m_size = dataSize;
}

void
TcpWebClient::SetFill (uint8_t fill, uint32_t dataSize)
{
  if (dataSize != m_dataSize)
    {
      delete [] m_data;
      m_data = new uint8_t [dataSize];
      m_dataSize = dataSize;
    }

  memset (m_data, fill, dataSize);

  //
  // Overwrite packet size attribute.
  //
  m_size = dataSize;
}

void
TcpWebClient::SetFill (uint8_t *fill, uint32_t fillSize, uint32_t dataSize)
{
  if (dataSize != m_dataSize)
    {
      delete [] m_data;
      m_data = new uint8_t [dataSize];
      m_dataSize = dataSize;
    }

  if (fillSize >= dataSize)
    {
      memcpy (m_data, fill, dataSize);
      return;
    }

  //
  // Do all but the final fill.
  //
  uint32_t filled = 0;
  while (filled + fillSize < dataSize)
    {
      memcpy (&m_data[filled], fill, fillSize);
      filled += fillSize;
    }

  //
  // Last fill may be partial
  //
  memcpy (&m_data[filled], fill, dataSize - filled);

  //
  // Overwrite packet size attribute.
  //
  m_size = dataSize;
}

void
TcpWebClient::ScheduleTransmit (Time dt)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_sendEvent = Simulator::Schedule (dt, &TcpWebClient::Send, this);
}

void
TcpWebClient::Send (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  NS_ASSERT (m_sendEvent.IsExpired ());

  Ptr<Packet> p;
  if (m_dataSize)
    {
      //
      // If m_dataSize is non-zero, we have a data buffer of the same size that we
      // are expected to copy and send.  This state of affairs is created if one of
      // the Fill functions is called.  In this case, m_size must have been set
      // to agree with m_dataSize
      //
      NS_ASSERT_MSG (m_dataSize == m_size, "TcpWebClient::Send(): m_size and m_dataSize inconsistent");
      NS_ASSERT_MSG (m_data, "TcpWebClient::Send(): m_dataSize but no m_data");
      p = Create<Packet> (m_data, m_dataSize);
      m_bytesSent += m_dataSize;
    }
  else
    {
      //
      // If m_dataSize is zero, the client has indicated that she doesn't care
      // about the data itself either by specifying the data size by setting
      // the corresponding atribute or by not calling a SetFill function.  In
      // this case, we don't worry about it either.  But we do allow m_size
      // to have a value different from the (zero) m_dataSize.
      //
      p = Create<Packet> (m_size);
      m_bytesSent += m_size;
    }
  // call to the trace sinks before the packet is actually sent,
  // so that tags added to the packet can be sent as well
  m_txTrace (p);
  m_socket->Send (p);

  ++m_sent;

  NS_LOG_INFO ("Client: sent " << m_size << " bytes to " << (AddressPrinter)m_peerAddress);

  if (m_sent < m_count)
    {
      ScheduleTransmit (m_interval);
    }
}

void
TcpWebClient::ReceivePacket (Ptr<Socket> socket)
{
  NS_LOG_FUNCTION (this << socket);
  Ptr<Packet> packet;
  Address from;

  while (packet = socket->RecvFrom (from))
    {
      m_recvBack++;
      m_bytesRecvBack += packet->GetSize ();

      NS_LOG_INFO ("Client: received Packet " <<m_recvBack << " with "<< packet->GetSize () << " bytes from " << (AddressPrinter)from << "(total received " << m_bytesRecvBack << ")" );

    }

    // Client uses HTTP 1.1 and keepalive is set to infinity
    //socket->Close();
    //m_socket->SetRecvCallback (MakeNullCallback<void, Ptr<Socket> > ());
    //socket = 0;
   
}

} // Namespace ns3
