/* tcp-web-server.cc
// TE816 - Modelagem e Avaliação de Desempenho
// Pós Graduação em Engenharia Elétrica
// Universidade Federal do Paraná
// This souce was developed as basis for homework 
//
// -- 10/11/2018
// ==================================
// Prof. Carlos Marcelo Pedroso
// pedroso@eletrica.ufpr.br
// http://www.eletrica.ufpr.br/pedroso
// Universidade Federal do Paraná
// Departamento de Engenharia Elétrica
// Centro Politécnico
// CEP: 81531-990 - Curitiba - PR - Brasil
 */

#include "ns3/log.h"
#include "ns3/double.h"
#include "ns3/ipv4-address.h"
#include "ns3/ipv6-address.h"
#include "ns3/address-utils.h"
#include "ns3/nstime.h"
#include "ns3/inet-socket-address.h"
#include "ns3/inet6-socket-address.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/uinteger.h"
#include "ns3/random-variable-stream.h"

#include "tcp-web-server.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE ("TcpWebServerApplication");
NS_OBJECT_ENSURE_REGISTERED (TcpWebServer);

TypeId
TcpWebServer::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::TcpWebServer")
    .SetParent<Application> ()
    .AddConstructor<TcpWebServer> ()
    .AddAttribute ("Port", "Port on which we listen for incoming connections.",
                   UintegerValue (80),
                   MakeUintegerAccessor (&TcpWebServer::m_port),
                   MakeUintegerChecker<uint16_t> ())
    .AddAttribute ("PageSize", "Size of request to be sent do server",
                   UintegerValue (100),
                   MakeUintegerAccessor (&TcpWebServer::SetDataSize,
                                         &TcpWebServer::GetDataSize),
                   MakeUintegerChecker<uint32_t> ())
  ;
  return tid;
}

TcpWebServer::TcpWebServer ()
{
  NS_LOG_FUNCTION_NOARGS ();
  m_socket = 0;
  m_socket6 = 0;
  m_running = false;
  m_avgPageSize=1000; //!
}

TcpWebServer::~TcpWebServer()
{
  NS_LOG_FUNCTION_NOARGS ();
  m_socket = 0;
  m_socket6 = 0;
}

void
TcpWebServer::SetDataSize (uint32_t avgPageSize)
{
  NS_LOG_FUNCTION (avgPageSize);

  //
  // If the client is setting the web page size this way, 
  // he doesn't care about the contents of the packet at all, so
  // neither will we.
  //
    m_avgPageSize = avgPageSize;
}

uint32_t
TcpWebServer::GetDataSize (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_avgPageSize;
}


void
TcpWebServer::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  Application::DoDispose ();
}

void
TcpWebServer::StartApplication (void)
{
  NS_LOG_FUNCTION_NOARGS ();

  m_running = true;

  if (m_socket == 0)
    {
      TypeId tid = TypeId::LookupByName ("ns3::TcpSocketFactory");
      m_socket = Socket::CreateSocket (GetNode (), tid);
      InetSocketAddress listenAddress = InetSocketAddress (Ipv4Address::GetAny (), m_port);
      m_socket->Bind (listenAddress);
      m_socket->Listen();
    }
  if (m_socket6 == 0)
    {
      TypeId tid = TypeId::LookupByName ("ns3::TcpSocketFactory");
      m_socket6 = Socket::CreateSocket (GetNode (), tid);
      Inet6SocketAddress listenAddress = Inet6SocketAddress (Ipv6Address::GetAny (), m_port);
      m_socket6->Bind (listenAddress);
      m_socket6->Listen();
    }

  m_socket->SetAcceptCallback (
        MakeNullCallback<bool, Ptr<Socket>, const Address &> (),
        MakeCallback (&TcpWebServer::HandleAccept, this));
  m_socket6->SetAcceptCallback (
        MakeNullCallback<bool, Ptr<Socket>, const Address &> (),
        MakeCallback (&TcpWebServer::HandleAccept, this));
}

void
TcpWebServer::StopApplication ()
{
  NS_LOG_FUNCTION_NOARGS ();

  m_running = false;

  if (m_socket != 0)
    {
      m_socket->Close ();
      m_socket->SetAcceptCallback (
            MakeNullCallback<bool, Ptr<Socket>, const Address &> (),
            MakeNullCallback<void, Ptr<Socket>, const Address &> () );
    }
  if (m_socket6 != 0)
    {
      m_socket6->Close ();
      m_socket6->SetAcceptCallback (
            MakeNullCallback<bool, Ptr<Socket>, const Address &> (),
            MakeNullCallback<void, Ptr<Socket>, const Address &> () );
    }
}

void
TcpWebServer::ReceivePacket (Ptr<Socket> s)
{
  NS_LOG_FUNCTION (this << s);

  Ptr<Packet> recvPacket;
  Ptr<Packet> responsePacket; 

  uint8_t *dataPtr;
  uint32_t pageSize;

  // Now we will generate a exponentially distributed number, and use as page size.
  Ptr<ExponentialRandomVariable> ev = CreateObject<ExponentialRandomVariable> ();
  ev->SetAttribute ("Bound",DoubleValue (1000000000));
  ev->SetAttribute ("Mean", DoubleValue ((double)m_avgPageSize));  


  pageSize= ev->GetInteger();

  dataPtr= new uint8_t [pageSize];
  memset(dataPtr,'A',pageSize);  

  Address from;

  while (recvPacket = s->RecvFrom (from))
    {
      if (recvPacket->GetSize () > 0)
        {
    	  NS_LOG_INFO ("Server: received " << recvPacket->GetSize () << " bytes from " << (AddressPrinter)from);

    	  recvPacket->RemoveAllPacketTags ();
    	  recvPacket->RemoveAllByteTags ();

    	  NS_LOG_INFO ("Server: request received ... sending answer with "<< pageSize <<" bytes back to client");

          responsePacket = Create<Packet> (dataPtr, pageSize);  

    	  s->Send (responsePacket);
        }
    }
}

void TcpWebServer::HandleAccept (Ptr<Socket> s, const Address& from)
{
  NS_LOG_FUNCTION (this << s << from);
  s->SetRecvCallback (MakeCallback (&TcpWebServer::ReceivePacket, this));
  s->SetCloseCallbacks(MakeCallback (&TcpWebServer::HandleSuccessClose, this),
  MakeNullCallback<void, Ptr<Socket> > () );
}

void TcpWebServer::HandleSuccessClose(Ptr<Socket> s)
{
  NS_LOG_FUNCTION (this << s);
  NS_LOG_LOGIC ("Client close received");
  s->Close();
  s->SetRecvCallback (MakeNullCallback<void, Ptr<Socket> > () );
  s->SetCloseCallbacks(MakeNullCallback<void, Ptr<Socket> > (),
      MakeNullCallback<void, Ptr<Socket> > () );
}

} // Namespace ns3
